unit DecimalCalculations;

interface

uses Windows {$IFDEF KOL}, KOL {$IFDEF DECIMAL_EXCEPTIONS}, err {$ENDIF}
             {$ELSE}, SysUtils {$ENDIF};

const MaxDigits = 37;

type
  TDecimal = packed record
    Exponent: SmallInt;
    Negative: Boolean;
    Mantissa: array[ 0..MaxDigits-1 ] of Byte;
  end;
  PDecimal = ^TDecimal;

function Int2Decimal( i: Integer ): TDecimal;
function Decimal2Int( const d: TDecimal ): Integer;
function Float2Decimal( E: Extended ): TDecimal;
function Decimal2Float( const d1: TDecimal ): Extended;
function Str2Decimal( const s: String ): TDecimal;
function Decimal2Str( const d1: TDecimal; MaxFixedWidth: Integer = MaxDigits + 2 ): String;
function DecimalFormat( const fmt: String; const d1: TDecimal ): String;
function IsDecimalZero( const d1: TDecimal ): Boolean;
function DecimalCompare( const d1, d2: TDecimal ): Integer;
function DecimalCompare0( const d1: TDecimal ): Integer;

function DecimalAdd( const d1, d2: TDecimal ): TDecimal;
function DecimalSub( const d1, d2: TDecimal ): TDecimal;
function NegateDecimal( const d1: TDecimal ): TDecimal;
function DecimalMul( const d1, d2: TDecimal ): TDecimal;
function DecimalDiv( const d1, d2: TDecimal ): TDecimal;
function DecimalRound( const d1: TDecimal; LeaveDigits: Integer ): TDecimal;
function DecimalTrunc( const d1: TDecimal; LeaveDigits: Integer ): TDecimal;
function DecimalRoot( const d1: TDecimal ): TDecimal;
function DecimalSquare( const d1: TDecimal ): TDecimal;


implementation

{$IFNDEF KOL}
function StrRepeat( const S: String; Count: Integer ): String;
var I, L: Integer;
begin
  L := Length( S );
  SetLength( Result, L * Count );
  for I := 0 to Count-1 do
    Move( S[ 1 ], Result[ 1 + I * L ], L );
end;
{$ENDIF}

type
  TDecimalDoubleSize = packed record
    MantissaD: array[ 0..MaxDigits*2-1 ] of Byte;
  end;

function Int2Decimal( i: Integer ): TDecimal;
var P: PByte;
    digits: Integer;
begin
  Result.Negative := i < 0;
  if Result.Negative then
    i := -i;
  digits := 0;
  P := Pointer( Integer( @ Result.Mantissa[ 0 ] ) + MaxDigits );
  while i > 0  do
  begin
    dec( P );
    P^ := i mod 10;
    inc( digits );
    i := i div 10;
  end;
  Result.Exponent := digits;
  if digits > 0 then
    move( P^, Result.Mantissa, digits );
  if digits < MaxDigits then
    FillChar( Result.Mantissa[ digits ], MaxDigits - digits, 0 );
end;

function Decimal2Int( const d: TDecimal ): Integer;
var P: PByte;
    digits: Integer;
begin
  Result := 0;
  digits := d.Exponent;
  if digits <= 0 then Exit;
  P := @ d.Mantissa;
  while digits > 0 do
  begin
    Result := Result * 10 + P^;
    inc( P );
    dec( digits );
  end;
end;

procedure Denormalize( const d1, d2: TDecimal; var p1, p2: PDecimal );
var dLE, dGE: PDecimal;
    count: Integer;
begin
  if d1.Exponent < d2.Exponent then
  begin
    p1^ := d1; p2 := @ d2;
    dLE := p1; dGE := p2;
  end
    else
  begin
    p1 := @ d1; p2^ := d2;
    dLE := p2; dGE := p1;
  end;
  count := dGE.Exponent - dLE.Exponent;
  if count >= MaxDigits then
    FillChar( dLE.Mantissa[ 0 ], MaxDigits, 0 )
  else
  begin
    Move( dLE.Mantissa[ 0 ], dLE.Mantissa[ count ], MaxDigits - count );
    FillChar( dLE.Mantissa[ 0 ], count, 0 );
  end;
  dLE.Exponent := dGE.Exponent;
end;

procedure Normalize( var d1: TDecimal );
var i, count: Integer;
    P: PByte;
begin
  count := 0;
  p := @ d1.Mantissa[ 0 ];
  for i := 0 to MaxDigits-1 do
  begin
    if p^ <> 0 then break;
    inc( count );
    inc( p );
  end;
  if count >= MaxDigits then
  begin
    d1.Exponent := 0;
    d1.Negative := FALSE;
  end
    else
  begin
    move( d1.Mantissa[ count ], d1.Mantissa[ 0 ], MaxDigits - count );
    FillChar( d1.Mantissa[ MaxDigits - count ], count, 0 );
    dec( d1.Exponent, count );
  end;
end;

function Float2Decimal( E: Extended ): TDecimal;
var mult, d256: TDecimal;
    F: Extended;
    Negative: Boolean;
    position: Integer;
    t: Integer;
begin
  mult := Int2Decimal( 1 );
  d256 := Int2Decimal( 256 );
  Negative := E < 0;
  if Negative then
    E := -E;
  while E > $40000000 do
  begin
    E := E / 256;
    mult := DecimalMul( mult, d256 );
  end;
  F := Frac( E );
  Result := Int2Decimal( Round( E - F ) );
  position := Result.Exponent;
  while (F > 0) and (position < MaxDigits) do
  begin
    F := F * 10;
    t := Trunc( F );
    Result.Mantissa[ position ] := t;
    F := F - t;
    inc( position );
  end;
  if Result.Mantissa[ 0 ] = 0 then
    Normalize( Result );
  Result.Negative := Negative;
end;

function Decimal2Float( const d1: TDecimal ): Extended;
var i: Integer;
    k: Extended;
begin
  Result := 0;
  k := 1;
  for i := d1.Exponent-1 downto 0 do
  begin
    if i < MaxDigits then
      Result := Result + d1.Mantissa[ i ] * k;
    k := k * 10;
  end;
  k := 1;
  for i := d1.Exponent to MaxDigits-1 do
  begin
    k := k * 0.1;
    if i >= 0 then
      Result := Result + d1.Mantissa[ i ] * k;
  end;
  if d1.Negative then Result := -Result;
end;

function Str2Decimal( const s: String ): TDecimal;
var p: PChar;
    DecimalPt: Boolean;
    r: PByte;
    e: Integer;
    nege: Boolean;
begin
  FillChar( Result, Sizeof( Result ), 0 );
  if s = '' then Exit;
  p := PChar( s );
  if p^ = '-' then
  begin
    Result.Negative := TRUE;
    inc( p );
  end;
  DecimalPt := FALSE;
  r := @ Result.Mantissa[ 0 ];
  while p^ <> #0 do
  begin
    CASE p^ OF
    '.': if DecimalPt then Exit
         else begin
                DecimalPt := TRUE;
              end;
    '0'..'9':
      begin
        if Result.Exponent <= MaxDigits then
        begin
          r^ := Ord( p^ ) - Ord( '0' );
          inc( r );
        end;
        if not DecimalPt then
          inc( Result.Exponent );
      end;
    'e', 'E':
      begin
        e := 0;
        nege := FALSE;
        inc( p );
        if p^ = '-' then
        begin
          nege := TRUE;
          inc( p );
        end;
        while p^ in [ '0'..'9' ] do
        begin
          e := e * 10 + Ord( p^ ) - Ord( '0' );
          inc( p );
        end;
        if nege then e := -e;
        inc( Result.Exponent, e );
        break;
      end;
    END;
    inc( p );
  end;
  if Result.Mantissa[ 0 ] = 0 then
    Normalize( Result );
end;

function Decimal2Str( const d1: TDecimal; MaxFixedWidth: Integer = MaxDigits + 2 ): String;
var CountTailZeroes, i, n: Integer;
    p: PByte;
    s, pe: PChar;
    e: String;
begin
  CountTailZeroes := 0;
  p := @ d1.Mantissa[ MaxDigits-1 ];
  for i := MaxDigits-1 downto 0 do
  begin
    if p^ = 0 then inc( CountTailZeroes )
    else break;
    dec( p );
  end;
  i := MaxDigits - CountTailZeroes;
  if d1.Exponent > i then
    inc( i, d1.Exponent - i )
  else
  if d1.Exponent <= 0 then
  begin
    inc( i, 1 - d1.Exponent );
    if i > 1 then inc( i );
  end
  else
  if d1.Exponent < i then
    inc( i ); //    !
  inc( i, Integer( d1.Negative ) );
  if i <= MaxFixedWidth then
  begin // fixed format
    SetLength( Result, i );
    s := @ Result[ 1 ];
    s[ i ] := #0;
    if d1.Negative then
    begin
      s^ := '-'; inc( s );
    end;
    if (d1.Exponent <= 0) then
    begin
      s^ := '0'; inc( s );
      if CountTailZeroes >= MaxDigits then Exit;
      s^ := '.'; inc( s );
      for i := 0 to -d1.Exponent-1 do
      begin
        s^ := '0'; inc( s );
      end;
    end;
    n := d1.Exponent;
    p := @ d1.Mantissa[ 0 ];
    for i := MaxDigits - CountTailZeroes - 1 downto 0 do
    begin
      s^ := Char( p^ + Ord( '0' ) );
      inc( s ); inc( p );
      dec( n );
      if n = 0 then
      begin
        if i = 0 then Exit;
        s^ := '.'; inc( s );
      end;
    end;
    while n > 0 do
    begin
      s^ := '0'; inc( s );
      dec( n );
    end;
  end
    else
  begin // floating, normalized
    e := {$IFDEF KOL} Int2Str( d1.Exponent ) {$ELSE} IntToStr( d1.Exponent ) {$ENDIF};
    i := MaxDigits - CountTailZeroes;
    if i < 2 then i := 2;
    SetLength( Result, i + 2 + Length( e ) + Integer( d1.Negative ) );
    s := @ Result[ 1 ];
    s[ Length( s ) ] := #0;
    if d1.Negative then
    begin
      s^ := '-'; inc( s );
    end;
    p := @ d1.Mantissa[ 0 ];
    s^ := Char( p^ + Ord( '0' ) ); inc( s ); inc( p );
    s^ := '.'; inc( s );
    s^ := Char( p^ + Ord( '0' ) ); inc( s ); inc( p );
    for i := 3 to MaxDigits - CountTailZeroes do
    begin
      s^ := Char( p^ + Ord( '0' ) ); inc( s );
      inc( p );
    end;
    s^ := 'E'; inc( s );
    pe := @ e[ 1 ];
    while pe^ <> #0 do
    begin
      s^ := pe^; inc( s ); inc( pe );
    end;
  end;
end;

function ScanInteger( var s: PChar ): Integer;
var neg: Boolean;
begin
  Result := 0;
  neg := s^ = '-';
  if neg then inc( s );
  while s^ in [ '0'..'9' ] do
  begin
    Result := Result * 10 + Ord( s^ ) - Ord( '0' );
    inc( s );
  end;
  if neg then Result := - Result;
end;

function DecimalFormat( const fmt: String; const d1: TDecimal ): String;
var s: PChar;
    w, f, e: Integer;
    {pt,} sg, zz, ee: Boolean;
    i: Integer;
    tmp: String;
    function AdjustFixedFormat( var x: String; w, f: Integer{; pt: Boolean} ): String;
    var i: Integer;
    begin
      Result := x;
      i := pos( '.', Result );
      if i <= 0 then i := Length( Result );
      if Result[1] in [ '-', '+' ] then dec( i ); // i =    
      if (w > 0) and (i < w) then
      begin //  (w-i)       
        if zz then //   ( )
          if Result[1] in [ '-', '+' ] then
            Result := Result[1] + StrRepeat( '0', w - i ) + Copy( Result, 2, MaxInt )
          else
            Result := StrRepeat( '0', w - i ) + Result
        else
          Result := StrRepeat( ' ', w - i ) + Result;
      end;
      if f > 0 then
      begin
        {i := pos( '.', Result );
        if (i <= 0) and pt then Result := Result + '.';}
        i := pos( '.', Result );
        if (i <= 0) then Result := Result + '.';
        i := pos( '.', Result );
        if Length( Result ) - i < f then
          Result := Result + StrRepeat( '0', f - (Length( Result ) - i) )
        else
        if Length( Result ) - i > f then
          Result := Copy( Result, 1, i + f );
      end;
    end;
begin
  if fmt = '' then
  begin
    Result := Decimal2Str( d1 );
    Exit;
  end;
  s := @ fmt[ 1 ];
  if s^ = '%' then inc( s );
  w := -1;
  f := -1;
  e := 0;
  //pt := FALSE;
  sg := FALSE;
  zz := FALSE;
  ee := FALSE;
  if s^ in [ '+', '-' ] then begin sg := TRUE; inc( s ); end;
  if s^ in [ '0'..'9' ] then
  begin
    zz := s^='0';
    w := ScanInteger( s );
  end;
  if s^ = '.' then
  begin
    {pt := TRUE;} inc( s );
    if s^ in [ '0'..'9' ] then f := ScanInteger( s );
  end;
  if s^ in [ 'e', 'E' ] then
  begin
    //inc( s );
    e := ScanInteger( s );
    ee := TRUE;
  end;
  if not ee then
  begin //    
    i := MaxDigits + 2;
    if d1.Exponent < 0 then dec( i, d1.Exponent );
    Result := Decimal2Str( d1, i );
    if sg and not d1.Negative then Result := '+' + Result;
    Result := AdjustFixedFormat( Result, w, f {, pt} );
  end
    else
  begin //   ..
    Result := Decimal2Str( d1, 0 );
    if sg and not d1.Negative then Result := '+' + Result;
    if (w > 0) or (f > 0) or (e > 0) then
    begin
      i := pos( 'E', Result );
      tmp := Copy( Result, 1, i-1 );
      Result := Copy( Result, i, MaxInt );
      tmp := AdjustFixedFormat( tmp, w, f {, FALSE} );
      if e > 0 then
      begin
        if e > Length( Result ) then
          Result := Copy( Result, 1, 2 ) + StrRepeat( '0', e - Length( Result ) ) +
                    Copy( Result, 3, MaxInt );
      end;
      Result := tmp + Result;
    end;
  end;
end;

function IsDecimalZero( const d1: TDecimal ): Boolean;
begin
  Result := (d1.Exponent=0) and (d1.Mantissa[0]=0);
end;

function DecimalCompare( const d1, d2: TDecimal ): Integer;
var t: TDecimal;
begin
  t := DecimalSub( d1, d2 );
  Result := DecimalCompare0( t );
end;

function DecimalCompare0( const d1: TDecimal ): Integer;
begin
  if d1.Negative then Result := -1
  else Result := d1.Mantissa[ 0 ];
end;

function CompareMantissa( pD1, pD2: PByte; count: Integer ): Integer;
begin
  Result := 0;
  while count > 0 do
  begin
    Result := Integer( pD1^ ) - Integer( pD2^ );
    if Result <> 0 then Exit;
    inc( pD1 );
    inc( pD2 );
    dec( count );
  end;
end;

function AddMantissa( pD1, pD2, pD3: PByte; count: Integer ): Boolean;
var i: Integer;
    Carry: Boolean;
    digit: Integer;
begin
  Carry := FALSE;
  for i := count-1 downto 0 do
  begin
    digit := Integer( Carry ) + pD1^ + pD2^;
    if digit >= 10 then
    begin
      Carry := TRUE;
      digit := digit - 10;
    end
      else Carry := FALSE;
    pD3^ := digit;
    dec( pD1 );
    dec( pD2 );
    dec( pD3 );
  end;
  Result := Carry;
end;

procedure SubtractMantissa( pD1, pD2, pD3: PByte; count: Integer );
var Carry: Boolean;
    digit: Integer;
begin
  Carry := FALSE;
  while count > 0 do
  begin
    digit := Integer( pD1^ ) - ( pD2^ ) - Integer( Carry );
    if digit < 0 then
    begin
      Carry := TRUE;
      digit := digit + 10;
    end
      else Carry := FALSE;
    pD3^ := digit;
    dec( pD1 );
    dec( pD2 );
    dec( pD3 );
    dec( count );
  end;
end;

function DecimalAdd( const d1, d2: TDecimal ): TDecimal;
var t1, t2: TDecimal;
    p1, p2: PDecimal;
begin
  if d1.Exponent <> d2.Exponent then
  begin
    p1 := @ t1;
    p2 := @ t2;
    Denormalize( d1, d2, p1, p2 );
  end
  else
  begin
    p1 := @ d1;
    p2 := @ d2;
  end;
  Result.Exponent := p1.Exponent;
  Result.Negative := p1.Negative;
  if d1.Negative = d2.Negative then
  begin
    if AddMantissa( @ p1.Mantissa[ MaxDigits-1 ],
                    @ p2.Mantissa[ MaxDigits-1 ],
                    @ Result.Mantissa[ MaxDigits-1 ], MaxDigits ) then
    begin
      inc( Result.Exponent );
      Move( Result.Mantissa[ 0 ], Result.Mantissa[ 1 ], MaxDigits-1 );
      Result.Mantissa[ 0 ] := 1;
    end;
  end
  else
  begin
    if CompareMantissa( @ p1.Mantissa[0], @ p2.Mantissa[ 0 ], MaxDigits ) > 0 then
      SubtractMantissa( @ p1.Mantissa[MaxDigits-1], @ p2.Mantissa[MaxDigits-1],
                        @ Result.Mantissa[MaxDigits-1], MaxDigits )
    else
    begin
      Result.Negative := d2.Negative;
      SubtractMantissa( @ p2.Mantissa[MaxDigits-1], @ p1.Mantissa[MaxDigits-1],
                        @ Result.Mantissa[MaxDigits-1], MaxDigits );
    end;
  end;
  if Result.Mantissa[ 0 ] = 0 then
    Normalize( Result );
end;

function DecimalSub( const d1, d2: TDecimal ): TDecimal;
var t2: TDecimal;
begin
  t2 := NegateDecimal( d2 );
  Result := DecimalAdd( d1, t2 );
end;

function NegateDecimal( const d1: TDecimal ): TDecimal;
begin
  Result := d1;
  Result.Negative := not Result.Negative;
end;

function DecimalMul( const d1, d2: TDecimal ): TDecimal;
var dd: array[ 0..2] of TDecimalDoubleSize;
    i, c: Integer;
    p, f: PByte;
begin
  FillChar( dd, SizeOf( dd ), 0 );
  Move( d1.Mantissa[ 0 ], dd[1].MantissaD[ 0 ], MaxDigits );
  p := @ d2.Mantissa[ 0 ];
  f := @ dd[1].MantissaD[ MaxDigits-1 ];
  Result.Negative := d1.Negative xor d2.Negative;
  Result.Exponent := d1.Exponent + d2.Exponent + MaxDigits - 1;
  for i := MaxDigits-1 downto 0 do
  begin
    c := p^;
    while c > 0 do
    begin
      if AddMantissa( f, @ dd[2].MantissaD[ MaxDigits*2-1 ],
        @ dd[2].MantissaD[ MaxDigits*2-1 ], MaxDigits * 2 ) then
      begin
        inc( Result.Exponent );
        Move( dd[2].MantissaD[ 0 ], dd[2].MantissaD[ 1 ], MaxDigits*2-1 );
        dec( f );
      end;
      dec( c );
    end;
    inc( p );
    dec( f );
  end;
  c := 0;
  for i := 0 to MaxDigits-1 do
    if dd[2].MantissaD[ i ] = 0 then
      inc( c )
    else break;
  Move( dd[2].MantissaD[ c ], Result.Mantissa[ 0 ], MaxDigits );
  dec( Result.Exponent, c );
end;

function DecimalDiv( const d1, d2: TDecimal ): TDecimal;
var dd: array[ 0..2] of TDecimalDoubleSize;
    i, c: Integer;
    p, f, t, f0, t0: PByte;
begin
  FillChar( Result, SizeOf( Result ), 0 );
  if IsDecimalZero( d2 ) then
  begin
    {$IFDEF DECIMAL_EXCEPTIONS}
    {$IFDEF KOL}
      raise Exception.Create( e_ZeroDivide, 'Decimal Zerodivide' );
    {$ELSE}
      raise Exception.Create( 'Decimal ZeroDivide' );
    {$ENDIF}
    {$ENDIF}
    Exit;
  end;
  FillChar( dd, SizeOf( dd ), 0 );
  Move( d1.Mantissa[ 0 ], dd[1].MantissaD[ 0 ], MaxDigits );
  Move( d2.Mantissa[ 0 ], dd[2].MantissaD[ MaxDigits ], MaxDigits );
  p := @ Result.Mantissa[ 0 ];
  Result.Negative := d1.Negative xor d2.Negative;
  Result.Exponent := d1.Exponent - d2.Exponent + 1;
  f := @ dd[1].MantissaD[ MaxDigits-1 ];
  f0 := @ dd[1].MantissaD[ 0 ];
  t := @ dd[2].MantissaD[ MaxDigits*2-1 ];
  t0 := @ dd[2].MantissaD[ MaxDigits ];
  for i := 0 to MaxDigits-1 do
  begin
    c := 0;
    while CompareMantissa( f0, t0, MaxDigits ) >= 0 do
    begin
      inc( c );
      SubtractMantissa( f, t, f, MaxDigits );
    end;
    p^ := c;
    inc( p );
    dec( t );
    dec( t0 );
  end;
  if Result.Mantissa[ 0 ] = 0 then
    Normalize( Result );
end;

function DecimalRound( const d1: TDecimal; LeaveDigits: Integer ): TDecimal;
var t1: TDecimal;
begin
  Result := d1;
  inc( LeaveDigits, d1.Exponent );
  if (LeaveDigits >= 0) and (LeaveDigits < MaxDigits) then
  begin
    if d1.Mantissa[ LeaveDigits ] >= 5 then
    begin
      FillChar( t1, SizeOf( TDecimal ), 0 );
      t1.Mantissa[ LeaveDigits-1 ] := 1;
      t1.Exponent := d1.Exponent;
      t1.Negative := d1.Negative;
      Result := DecimalAdd( d1, t1 );
    end;
    FillChar( Result.Mantissa[ LeaveDigits ], MaxDigits - LeaveDigits, 0 );
  end
    else
  if LeaveDigits < 0 then
    FillChar( Result, SizeOf( Result ), 0 )
end;

function DecimalTrunc( const d1: TDecimal; LeaveDigits: Integer ): TDecimal;
begin
  Result := d1;
  inc( LeaveDigits, d1.Exponent );
  if (LeaveDigits >= 0) and (LeaveDigits < MaxDigits) then
    FillChar( Result.Mantissa[ LeaveDigits ], MaxDigits - LeaveDigits, 0 );
end;

const
  MaxRootKoefCount = MaxDigits * 10;
var
  RootKoefInitialized: Boolean;
  RootKoef: array[ 1..MaxRootKoefCount ] of TDecimal;

procedure InitializeRootKoef;
var nom, den, d2, k: TDecimal;
    i: Integer;
begin
  nom := Int2Decimal( 1 );
  den := Int2Decimal( 2 );
  d2 := den;
  RootKoef[ 1 ] := DecimalDiv( nom, den );
  nom := Int2Decimal( 1 );
  den := Int2Decimal( 8 );
  for i := 2 to MaxRootKoefCount do
  begin
    RootKoef[ i ] := DecimalDiv( nom, den );
    if i mod 2 = 0 then
       RootKoef[ i ] := NegateDecimal( RootKoef[ i ] );
    k := Int2Decimal( i*2-1 );
    nom := DecimalMul( nom, k );
    k := Int2Decimal( i*2+2 );
    den := DecimalMul( den, k );
  end;
  RootKoefInitialized := TRUE;
end;

function DecimalRoot_Tailor( const d1: TDecimal ): TDecimal;
var x, xx, tmp: TDecimal;
    i, addexp: Integer;
begin
  FillChar( Result, SizeOf( Result ), 0 );
  i := DecimalCompare( d1, Int2Decimal( 0 ) );
  if i < 0 then
  begin
    {$IFDEF DECIMAL_EXCEPTIONS}
    {$IFDEF KOL}
    raise Exception.Create( e_custom, 'Decimal Root from Negative number' );
    {$ELSE}
    raise Exception.Create( 'Decimal Root from Negative number' );
    {$ENDIF}
    {$ENDIF}
    Exit;
  end
    else
  if i = 0 then Exit;

  if not RootKoefInitialized then
    InitializeRootKoef;

  x := d1;
  addexp := 0;
  i := x.Exponent;
  if i >= 1 then
  begin
    addexp := (i+1) div 2;
    dec( x.Exponent, addexp * 2 );
  end;

  xx := Int2Decimal( -1 );
  x := DecimalAdd( x, xx );
  if DecimalCompare0( x ) = 0 then
    Result := Int2Decimal( 1 )
  else
  begin
    xx := x;
    Result := Int2Decimal( 1 );
    for i := 1 to MaxRootKoefCount do
    begin
      tmp := DecimalMul( xx, RootKoef[ i ] );
      if DecimalCompare0( tmp ) = 0 then break;
      Result := DecimalAdd( Result, tmp );
      if i < MaxDigits-1 then
        xx := DecimalMul( xx, x );
    end;
  end;
  inc( Result.Exponent, addexp );
end;

function DecimalRoot_FPU( const d1: TDecimal ): TDecimal;
var e: Extended;
begin
  e := Decimal2Float( d1 );
  e := sqrt( e );
  Result := Float2Decimal( e );
end;

function DecimalRoot( const d1: TDecimal ): TDecimal;
var i, d, probe: Integer;
    d2: TDecimal;
begin
  FillChar( Result, SizeOf( Result ), 0 );
  if d1.Negative then
  begin
    {$IFDEF DECIMAL_EXCEPTIONS}
    {$IFDEF KOL}
    raise Exception.Create( e_custom, 'Decimal root from negative number' );
    {$ELSE}
    raise Exception.Create( 'Decimal root from negative number' );
    {$ENDIF}
    {$ENDIF}
    Exit;
  end
    else
  if IsDecimalZero( d1 ) then Exit;
  Result.Exponent := (d1.Exponent + 1) div 2;
  for i := 0 to MaxDigits-1 do
  begin
    d := 1;
    probe := -1;
    while d < 10 do
    begin
      Result.Mantissa[ i ] := d;
      d2 := DecimalMul( Result, Result );
      probe := DecimalCompare( d2, d1 );
      if probe = 0 then break;
      if probe > 0 then
      begin
        Result.Mantissa[ i ] := d-1;
        break;
      end;
      inc( d );
    end;
    if probe = 0 then break;
  end;
  if Result.Mantissa[ 0 ] = 0 then
    Normalize( Result );
end;

function DecimalSquare( const d1: TDecimal ): TDecimal;
begin
  Result := DecimalMul( d1, d1 );
end;

end.
